import { defaults } from "../../utils/functions";
import * as getters from "./getters";
const RE_CONFIG = /([^:]+):([^,]*),?/g;
const RE_NUMBER = /^\-?\d*\.?\d+(?:e-?\d+)?$/;
export class Watcher {
    constructor(settings) {
        this.options = {
            element: document.body,
            prefix: "ob",
            components: {},
            getters: Object.assign({}, getters),
        };
        this.options = defaults(settings, this.options);
        this.instances = new Map();
    }
    runCreate(target, components, name) {
        let el = target instanceof Element ? target : target.$element;
        let parent = target instanceof Element ? undefined : target;
        for (let k in components) {
            let comp = components[k];
            let fullName = name ? `${name}-${k}` : k;
            let attr = `${this.options.prefix}-${fullName}`;
            let sel = `[${attr}]`;
            let qs = Array.from(el.querySelectorAll(sel));
            if (el.matches(sel)) {
                qs.unshift(el);
            }
            for (let i = 0, l = qs.length; i < l; i++) {
                let eq = qs[i];
                let comps = this.instances.get(eq);
                if (!comps) {
                    comps = new Map();
                    this.instances.set(eq, comps);
                }
                else if (comps.get(comp)) {
                    continue;
                }
                let inst;
                try {
                    inst = new comp(eq, this.parseOptions(eq.getAttribute(attr), eq));
                }
                catch (e) {
                    console.error(`Create ${fullName}:`, e);
                    continue;
                }
                comps.set(comp, inst);
                if (parent) {
                    parent.$addChild(inst);
                }
                if (comp.components) {
                    this.runCreate(inst, comp.components, fullName);
                }
                if (!parent) {
                    // Children are initialized by parents, so only parents
                    // should be initialized.
                    try {
                        inst.$init();
                    }
                    catch (e) {
                        console.error(`Init ${fullName}:`, e);
                    }
                }
            }
        }
    }
    runDestroy(target) {
        this.instances.forEach((comps, el) => {
            if (el === target || target.contains(el)) {
                comps.forEach((c) => c.$destroy());
                this.instances.delete(target);
            }
        });
    }
    handleMutations(list) {
        list.forEach((mutation) => {
            mutation.addedNodes.forEach((n) => {
                if (n instanceof Element) {
                    this.runCreate(n, this.options.components);
                }
            });
            mutation.removedNodes.forEach((n) => {
                if (n instanceof Element) {
                    this.runDestroy(n);
                }
            });
        });
    }
    parseOptions(value, element) {
        if (value[0] === "{") {
            return JSON.parse(value);
        }
        let result = {};
        let match;
        RE_CONFIG.lastIndex = 0;
        while ((match = RE_CONFIG.exec(value))) {
            let k = match[1].trim();
            let v = match[2].trim();
            let key = k;
            let val = v;
            if (v.match(RE_NUMBER)) {
                val = parseFloat(v);
            }
            else if (v === "true") {
                val = true;
            }
            else if (v === "false") {
                val = false;
            }
            else if (v === "null") {
                val = null;
            }
            else if (v === "undefined" || v.length === 0) {
                val = undefined;
            }
            if (k.indexOf("$") > 0) {
                let split = k.split("$");
                let getter = this.options.getters[split[1]];
                key = split[0];
                val = getter ? getter(val, element) : null;
            }
            result[key] = val;
        }
        return result;
    }
    run() {
        this.runCreate(this.options.element, this.options.components);
    }
    observe() {
        this.observer = new MutationObserver((list) => this.handleMutations(list));
        this.observer.observe(this.options.element, {
            childList: true,
            subtree: true,
        });
    }
    get(element, constructor) {
        var _a;
        return (_a = this.instances.get(element)) === null || _a === void 0 ? void 0 : _a.get(constructor);
    }
    destroy() {
        var _a;
        this.instances.clear();
        (_a = this.observer) === null || _a === void 0 ? void 0 : _a.disconnect();
    }
}
export default Watcher;
