/**
 * Returns the first parent element that matches the predicate.
 */
export function findAncestor(element, predicate) {
    while ((element = element.parentElement)) {
        if (predicate(element) === true) {
            return element;
        }
    }
    return null;
}
/**
 * Returns the first element in the parent tree that has an `<a>` tag, or the
 * input element itself.
 */
export function findAnchor(element) {
    if (element.tagName === "A") {
        return element;
    }
    else {
        return findAncestor(element, (e) => e.tagName === "A");
    }
}
/**
 * Returns the page offset of an element according to the sum of its ancestors'
 * offsets.
 */
export function offsetGlobal(element, reference = null) {
    let offset = { top: 0, left: 0 };
    do {
        offset.top += element.offsetTop;
        offset.left += element.offsetLeft;
        element = element.offsetParent;
    } while (element !== reference);
    return offset;
}
/**
 * Calls getBoundingClientRect when the input is an Element and returns it. If
 * the input is Window, an object with a ClientRect shape is returned instead.
 */
export function getClientRect(input) {
    if (input instanceof Element) {
        return input.getBoundingClientRect();
    }
    else if (input === window) {
        return new DOMRect(0, 0, window.innerWidth, window.innerHeight);
    }
}
/**
 * Uses the Selection API to put a string in the user's clipboard.
 */
export function copyToClipboard(input) {
    let selection = window.getSelection();
    let node = document.createTextNode(input);
    let range = document.createRange();
    selection.removeAllRanges();
    document.body.appendChild(node);
    range.selectNodeContents(node);
    selection.addRange(range);
    document.execCommand("copy");
    document.body.removeChild(node);
}
/**
 * Injects a `<script>` tag in the `<head>` and returns a Promise.
 */
export function injectScript(src, async = true) {
    return new Promise((resolve, reject) => {
        let script = document.createElement("script");
        script.addEventListener("load", resolve);
        script.addEventListener("error", reject);
        script.type = "text/javascript";
        script.async = async;
        script.src = src;
        document.head.appendChild(script);
    });
}
