import { ticker } from "../../core/ticker";
import { Emitter } from "../emitter";
import { Point, Vector } from "../math";
export class Swipe {
    constructor(id, origin) {
        this.id = id;
        this.origin = origin;
        this.position = origin.clone();
        this.positionPrev = null;
        this.delta = null;
        this.speed = null;
    }
    offset() {
        return this.position.clone().subPoint(this.origin);
    }
}
export class Gesture extends Emitter {
    constructor(element) {
        super();
        this.element = element;
        this.swipes = [];
        this.mouseStartHandler = this.mouseStart.bind(this);
        this.mouseMoveHandler = this.mouseMove.bind(this);
        this.mouseEndHandler = this.mouseEnd.bind(this);
        this.touchStartHandler = this.touchStart.bind(this);
        this.touchMoveHandler = this.touchMove.bind(this);
        this.touchEndHandler = this.touchEnd.bind(this);
        this.tickHandler = this.handleTick.bind(this);
        this.element.addEventListener("mousedown", this.mouseStartHandler);
        this.element.addEventListener("touchstart", this.touchStartHandler);
        this.element.addEventListener("touchmove", this.touchMoveHandler);
        this.element.addEventListener("touchend", this.touchEndHandler);
        this.element.addEventListener("touchcancel", this.touchEndHandler);
    }
    getEventPoint(input) {
        return new Point(input.clientX, input.clientY);
    }
    touchStart(event) {
        for (let touch of event.changedTouches) {
            let origin = this.getEventPoint(touch);
            let swipe = new Swipe(touch.identifier, origin);
            this.swipes.push(swipe);
        }
        this.emitStart(event);
    }
    touchMove(event) {
        for (let touch of event.changedTouches) {
            let swipe = this.swipes.find((el) => el.id === touch.identifier);
            if (swipe) {
                swipe.position = this.getEventPoint(touch);
            }
        }
        this.emitMove(event);
    }
    touchEnd(event) {
        for (let touch of event.changedTouches) {
            this.swipes = this.swipes.filter((el) => el.id !== touch.identifier);
        }
        this.emitEnd(event);
    }
    mouseStart(event) {
        let origin = this.getEventPoint(event);
        let swipe = new Swipe("mouse", origin);
        this.swipes.push(swipe);
        this.element.addEventListener("mousemove", this.mouseMoveHandler);
        this.element.addEventListener("mouseup", this.mouseEndHandler);
        this.element.addEventListener("mouseleave", this.mouseEndHandler);
        this.emitStart(event);
    }
    mouseMove(event) {
        let swipe = this.swipes.find((el) => el.id === "mouse");
        if (swipe) {
            swipe.position = this.getEventPoint(event);
        }
        this.emitMove(event);
    }
    mouseEnd(event) {
        this.element.removeEventListener("mousemove", this.mouseMoveHandler);
        this.element.removeEventListener("mouseup", this.mouseEndHandler);
        this.element.removeEventListener("mouseleave", this.mouseEndHandler);
        this.swipes = this.swipes.filter((el) => el.id !== "mouse");
        this.emitEnd(event);
    }
    emitStart(event) {
        if (this.swipes.length === 1) {
            this.emit("start", event);
        }
        this.emit("down", event);
        ticker.on("tick", this.tickHandler);
    }
    emitMove(event) {
        this.emit("move", event);
    }
    emitEnd(event) {
        if (this.swipes.length === 0) {
            this.emit("end", event);
        }
        this.emit("up", event);
        ticker.off("tick", this.tickHandler);
    }
    handleTick(delta) {
        this.swipes.forEach((el) => {
            if (el.positionPrev) {
                el.delta = new Vector(el.positionPrev, el.position);
                el.speed = el.delta.magnitude * (1000 / delta);
            }
            el.positionPrev = el.position;
        });
    }
    destroy() {
        this.element.removeEventListener("mousedown", this.mouseStartHandler);
        this.element.removeEventListener("mousemove", this.mouseMoveHandler);
        this.element.removeEventListener("mouseup", this.mouseEndHandler);
        this.element.removeEventListener("mouseleave", this.mouseEndHandler);
        this.element.removeEventListener("touchstart", this.touchStartHandler);
        this.element.removeEventListener("touchmove", this.touchMoveHandler);
        this.element.removeEventListener("touchend", this.touchEndHandler);
        this.element.removeEventListener("touchcancel", this.touchEndHandler);
        ticker.off("tick", this.tickHandler);
    }
}
